/* 
    bttv - Bt848 frame grabber driver

    Copyright (C) 1996,97 Ralph Metzler (rjkm@thp.uni-koeln.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/types.h>
#include <linux/time.h>
#include <linux/bios32.h>
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/segment.h>

#include <linux/version.h>
#if LINUX_VERSION_CODE >= 0x020100
#include <asm/uaccess.h>
#define LINUX_2_1
#else
#define copy_from_user memcpy_fromfs
#define copy_to_user memcpy_tofs 
#define ioremap vremap 
#define iounmap vfree
#endif

#include "bttv.h"
#include "tuner.h"


#include <X11/X.h>
#include <X11/Xlib.h>

static unsigned int remap=0;
static unsigned int vidmem=0;
static unsigned int tuner=TUNER_DEFAULT;
#if LINUX_VERSION_CODE >= 0x020117
MODULE_PARM(tuner,"i");
#endif

static int triton1=0;

static int find_vga(void);
static void bt848_set_risc_jmps(struct bttv *btv);
static int read_msp3400c(struct bttv *btv,int dev, int addr);
static void write_msp3400c(struct bttv *btv,int dev, int addr, int val);

/* Anybody who uses more than four? */
#define BTTV_MAX 4

static int bttv_num;
static struct bttv bttvs[BTTV_MAX];

#define I2C_TIMING (0x7<<4)
#define I2C_COMMAND (I2C_TIMING | BT848_I2C_SCL | BT848_I2C_SDA)
#define I2CDELAY 20

#define AUDIO_MUTE_DELAY 10000
#define FREQ_CHANGE_DELAY 20000
#define EEPROM_WRITE_DELAY 20000

static int major = BTTV_MAJOR;


/*******************************/
/* Memory management functions */
/*******************************/

/* convert virtual user memory address to physical address */
/* (virt_to_phys only works for kmalloced kernel memory) */

static inline ulong
uvirt_to_phys(ulong adr) {
  pgd_t *pgd;
  pmd_t *pmd;
  pte_t *ptep, pte;
  
  pgd = pgd_offset(current->mm, adr);
  if (pgd_none(*pgd))
    return 0;
  pmd = pmd_offset(pgd, adr);
  if (pmd_none(*pmd))
    return 0;
  ptep = pte_offset(pmd, adr/*&(~PGDIR_MASK)*/);
  pte = *ptep;
  if(pte_present(pte))
    return (pte_page(pte)|(adr&(PAGE_SIZE-1)));
  return 0;
}

static inline ulong
uvirt_to_bus(ulong adr) {
  return virt_to_bus(phys_to_virt(uvirt_to_phys(adr)));
}

/* convert virtual kernel memory (vmalloced mem) address to physical address */
static inline ulong
kvirt_to_phys(ulong adr) {
  return uvirt_to_phys(VMALLOC_VMADDR(adr));
}

static inline ulong
kvirt_to_bus(ulong adr) {
  return uvirt_to_bus(VMALLOC_VMADDR(adr));
}


/* WARNING! WARNING! WARNING!
   The following routine does NOT lock the memory, it just looks up
   the real (physical) addresses of user supplied memory which 
   should already be faulted and locked by the user and returns them in iov.
   The real locking routine with the same call parameters is currently in
   the ATM driver which will probably become part of kernels >=2.2.0
   It is not possible to implement this in a module and I do not want to
   get into kernel patching (for now).
*/
   
#if LINUX_VERSION_CODE < 0x020200
static int lock_user(unsigned long start, unsigned long size, int iov_max, struct iovec *iov)
{
  int iovcnt;
  int i;
  unsigned long pend, end, pmem, pstart;

  iovcnt = 0;
  end = start+size;

  if (end <= start)
    return 0;
  
  pstart = start & PAGE_MASK;
  pend = end & PAGE_MASK;

  if (!(pmem = uvirt_to_bus(start)))
    return -EINVAL;

  if (!iov_max) 
    return -ENOSPC;

  iov->iov_base = (caddr_t) pmem;

  if (pstart == pend) {
    iov->iov_len = size;
    return 1;
  }
  
  iov->iov_len = PAGE_SIZE-(start&(PAGE_SIZE-1));
  iovcnt++; 
  iov++;
  
  if (iovcnt == iov_max)
    return iovcnt;

  start = pstart+PAGE_SIZE;
  
  while (start<pend) 
   { 
    if (!(pmem = uvirt_to_bus(start)))
      return -EINVAL;

    if (iovcnt >= iov_max)
      return -ENOSPC;

    iov->iov_base = (caddr_t) pmem;
    iov->iov_len = PAGE_SIZE;
    
    iovcnt++; 
    iov++;
    
    start += PAGE_SIZE;
   }

  if (start<end) 
   {
    if (!(pmem = uvirt_to_bus(start)))
      return -EINVAL;

    if (iovcnt >= iov_max)
      return -ENOSPC;

    iov->iov_base = (caddr_t) pmem;
    iov->iov_len = end-start;
    
    iovcnt++; 
    iov++;
   }


#ifdef DEBUG
  printk("bttv: user_lock succeed\n");
#endif  
  return iovcnt;
}

static void unlock_user(int iovcnt, struct iovec *iov)
{

}
#endif


/*****************/
/* I2C functions */
/*****************/

static void 
I2CStart(struct bttv *btv)
{
  btwrite(1, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(3, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(2, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(0, BT848_I2C);
  udelay(I2CDELAY);
#ifdef DEBUG
  printk("< ");
#endif
}

static void 
I2CStop(struct bttv *btv)
{
  btwrite(0, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(2, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(3, BT848_I2C);
  udelay(I2CDELAY);
#ifdef DEBUG
  printk(">\n");
#endif
}

static void
I2COne(struct bttv *btv)
{
  btwrite(1, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(3, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(1, BT848_I2C);
  udelay(I2CDELAY);
}

static void
I2CZero(struct bttv *btv)
{
  btwrite(0, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(2, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(0, BT848_I2C);
  udelay(I2CDELAY);
}

static int
I2CAck(struct bttv *btv)
{
  int ack;

  btwrite(1, BT848_I2C);
  udelay(I2CDELAY);
  btwrite(3, BT848_I2C);
  udelay(I2CDELAY);
  ack=(btread(BT848_I2C)&1);
  btwrite(1, BT848_I2C);
  udelay(I2CDELAY);
  return ack;
}

static int
I2CSendByte(struct bttv *btv, unchar data, int wait_for_ack)
{
  int i, ack;

  btwrite(0, BT848_I2C);

  for (i=7; i>=0; i--) {
    if (data&(1<<i))
      I2COne(btv);
    else
      I2CZero(btv);
  }
  if (wait_for_ack)
    udelay(wait_for_ack);
  ack= I2CAck(btv);
#ifdef DEBUG
  printk("%02x%c ",(int)data,ack?'-':'+');
#endif
  return ack;
}

static unchar 
I2CReadByte(struct bttv *btv, int last)
{
  int i;
  unchar data=0;

  btwrite(1, BT848_I2C);
  udelay(I2CDELAY);
  for (i=7; i>=0; i--) {
    btwrite(3, BT848_I2C);
    udelay(I2CDELAY);
    if (btread(BT848_I2C)&1)
      data |= (1<<i);
    btwrite(1, BT848_I2C);
    udelay(I2CDELAY);
  }
  if (last)
    I2COne(btv);
  else
      I2CZero(btv);
#ifdef DEBUG
  printk("=3D%02x%c ",(int)data,last?'-':'+');
#endif
  return data;
}

static int
I2CAdrAck(struct bttv *btv, unchar adr) 
{
  int ack,i;

  for (i=0; i<5; i++) {
    I2CStart(btv);
    ack=I2CSendByte(btv, adr,0);
    I2CStop(btv);
    if (!ack)
      break;
  }
  return ack;
}

/* hardware I2C functions */

static int I2CRead(struct bttv *btv, int addr)
{
  u_long i;
  u_long stat;
  
  /* clear status bit ; BT848_INT_RACK is ro */
  btwrite(BT848_INT_I2CDONE, BT848_INT_STAT);
  
  btwrite(((addr & 0xff) << 24) | I2C_COMMAND, BT848_I2C);
  
  for (i=0x7fffffff; i; i--)
    {
      stat=btread(BT848_INT_STAT);
      if (stat & BT848_INT_I2CDONE)
	  break;
    }
  
  if (!i) return -1;
  if (!(stat & BT848_INT_RACK)) return -2;
  
  i=(btread(BT848_I2C)>>8)&0xff;
  return i;
}

static int I2CWrite(struct bttv *btv, unchar addr, unchar b1,
		    unchar b2, int both)
{
  u_long i;
  u_long data;
  u_long stat;
  
  /* clear status bit; BT848_INT_RACK is ro */
  btwrite(BT848_INT_I2CDONE, BT848_INT_STAT);
  
  data=((addr & 0xff) << 24) | ((b1 & 0xff) << 16) | I2C_COMMAND;
  if (both)
    {
      data|=((b2 & 0xff) << 8);
      data|=BT848_I2C_W3B;
    }
  
  btwrite(data, BT848_I2C);

  for (i=0x7fffffff; i; i--)
    {
      stat=btread(BT848_INT_STAT);
      if (stat & BT848_INT_I2CDONE)
	  break;
    }
  
  if (!i) return -1;
  if (!(stat & BT848_INT_RACK)) return -2;
  
  return 0;
}

static void readee(struct bttv *btv, unchar *eedata)
{
  int i, k;
  
  if (I2CWrite(btv, 0xa0, 0, -1, 0)<0)
    {
      printk("bttv: readee error\n");
      return;
    }

  for (i=0; i<256; i++)
    {
      k=I2CRead(btv, 0xa1);
      if (k<0)
	{
	  printk("bttv: readee error\n");
	  break;
	}
      eedata[i]=k;
    }
}

static void writeee(struct bttv *btv, unchar *eedata)
{
  int i;
  
  for (i=0; i<256; i++)
    {
      if (I2CWrite(btv, 0xa0, i, eedata[i], 1)<0)
	{
	  printk("bttv: writeee error (%d)\n", i);
	  break;
	}
      udelay(EEPROM_WRITE_DELAY);
    }
}

/* tuner, internal, external and mute */
static unchar audiomuxs[][4] = {
  { 0x00, 0x00, 0x00, 0x00}, /* unknown */
  { 0x02, 0x00, 0x00, 0x0a}, /* MIRO */
  { 0x00, 0x02, 0x03, 0x04}, /* Hauppauge */
  { 0x04, 0x02, 0x03, 0x01}, /* STB */
  { 0x01, 0x02, 0x03, 0x04}, /* Intel??? */
  { 0x01, 0x02, 0x03, 0x04}, /* Diamond DTV2000??? */
};

static void
audio(struct bttv *btv, int mode)
{
  btwrite(0x0f, BT848_GPIO_OUT_EN);
  btwrite(0x00, BT848_GPIO_REG_INP);

  switch (mode) {
  case AUDIO_UNMUTE:
    btv->audio&=~AUDIO_MUTE;
    mode=btv->audio;
    break;

  case AUDIO_OFF:
    mode=AUDIO_OFF;
    break;

  case AUDIO_ON:
    mode=btv->audio;
    break;

  default:
    btv->audio&=AUDIO_MUTE;
    btv->audio|=mode;
    break;
  }
  if ((btv->audio&AUDIO_MUTE) || !(btread(BT848_DSTATUS)&BT848_DSTATUS_HLOC))
    mode=AUDIO_OFF;
  btaor(audiomuxs[btv->type][mode] , ~0x0f, BT848_GPIO_DATA);
}


inline static void
bt848_dma(struct bttv *btv, uint state)
{
  if (state)
    btor(3, BT848_GPIO_DMA_CTL);
  else
    btand(~3, BT848_GPIO_DMA_CTL);
}


static void 
bt848_cap(struct bttv *btv, uint state)
{
  if (state) {
#ifdef DEBUG
    printk ("bttv: cap_on()\n");
#endif
    btv->cap|=3;
    bt848_set_risc_jmps(btv);
   } else {
#ifdef DEBUG
    printk ("bttv: cap_off()\n");
#endif
    btv->cap&=~3;
    bt848_set_risc_jmps(btv);
  }
}

static void 
bt848_muxsel(struct bttv *btv, uint input)
{
  input&=3;

  /* This seems to get rid of some synchronization problems */
  btand(~(3<<5), BT848_IFORM);
  udelay(10000); 

  if (input==3) {
    btor(BT848_CONTROL_COMP, BT848_E_CONTROL);
    btor(BT848_CONTROL_COMP, BT848_O_CONTROL);
  } else {
    btand(~BT848_CONTROL_COMP, BT848_E_CONTROL);
    btand(~BT848_CONTROL_COMP, BT848_O_CONTROL);
  }
  if (input==2)
    input=3;
  btaor(((input+2)&3)<<5, ~(3<<5), BT848_IFORM);
  audio(btv, input ? AUDIO_EXTERN : AUDIO_TUNER);
}

static void
make_vbitab(struct bttv *btv)
{
  int i;
  dword *po=(dword *) btv->vbi_odd;
  dword *pe=(dword *) btv->vbi_even;
  
  *(po++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1; *(po++)=0;
  for (i=0; i<16; i++) {
    *(po++)=BT848_RISC_WRITE|2044|BT848_RISC_EOL|BT848_RISC_SOL|(13<<20);
    *(po++)=kvirt_to_bus((ulong)btv->vbibuf+i*2048);
  }
  *(po++)=BT848_RISC_JUMP;
  *(po++)=virt_to_bus(btv->risc_jmp+4);

  *(pe++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1; *(pe++)=0;
  for (i=16; i<32; i++) {
    *(pe++)=BT848_RISC_WRITE|2044|BT848_RISC_EOL|BT848_RISC_SOL;
    *(pe++)=kvirt_to_bus((ulong)btv->vbibuf+i*2048);
  }
  *(pe++)=BT848_RISC_JUMP|BT848_RISC_IRQ|(0x01<<16);
  *(pe++)=virt_to_bus(btv->risc_jmp+10);
}

static int 
make_iov_risctab(struct bttv *btv, dword *ro, dword *re,
		 int iov_num, struct iovec *iov)
{
  ulong i;
  ulong bpl;  /* bytes per line */
  dword iovp; 
  ulong iovl;
  ulong bl;
  dword **rp;

  bpl=btv->win.width*btv->win.bpp;

  *(ro++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1|(14<<20); *(ro++)=0;
  *(re++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1; *(re++)=0;
  iovl=iov->iov_len;
  iovp=(dword) iov->iov_base;
  
  /* Note that TV line numbering starts with 1 but we start with 0 here.
     So, for even i we need to output odd lines! */
  for (i=0; i < (btv->win.height<<(1^btv->win.interlace)); i++) {
    if (btv->win.interlace) 
      rp= (i&1) ? &re : &ro;
    else 
      rp= (i>btv->win.height) ? &re : &ro; 

    if (bpl<=iovl) {
      *((*rp)++)=BT848_RISC_WRITE|BT848_RISC_SOL|BT848_RISC_EOL|bpl; 
      *((*rp)++)=iovp;
      iovp+=bpl; iovl-=bpl;
      if (!iovl) {
	iov++;
	iovl=iov->iov_len;
	iovp=(dword) iov->iov_base;
      }
    } else {
      *((*rp)++)=BT848_RISC_WRITE|BT848_RISC_SOL|iovl;
      *((*rp)++)=iovp;
      bl=bpl-iovl;
      iov++;
      while (bl>iov->iov_len) {
	*((*rp)++)=BT848_RISC_WRITE|iov->iov_len;
	*((*rp)++)=(dword) iov->iov_base;
	bl-=iov->iov_len;
	iov++;
      }
      *((*rp)++)=BT848_RISC_WRITE|BT848_RISC_EOL|bl;
      *((*rp)++)=(dword) iov->iov_base;
      iovl=iov->iov_len-bl; iovp=bl+(dword)iov->iov_base;      
    }
  }

  *(ro++)=BT848_RISC_JUMP;
  *(ro++)=btv->bus_vbi_even;
  *(re++)=BT848_RISC_JUMP|BT848_RISC_IRQ|(2<<16);
  *(re++)=btv->bus_vbi_odd;

  return 0;
}


static inline void
write_risc_virt(dword **rp, int opnum, dword *ops)
{
  dword *p,i;
  
  p=(dword *) (PAGE_ALIGN((dword)(*rp)));
  if ((*rp)+opnum+2>=p) {
    *((*rp)++)=BT848_RISC_JUMP;
    *((*rp)++)=kvirt_to_bus((dword)p);
    *rp=p;
  }
  p=(dword *) kvirt_to_bus((dword)(*rp));
  for (i=0; i<opnum; i++)
    *(p++)=ops[i];
  *rp+=opnum;
}


/* does this really make a difference ???? */
#define BURST_MAX 4096

static void
write_risc_segment(dword **rp, ulong line_adr, dword command,
		uint x, uint dx, uint bpp, uint width)
{
  dword flags, len;
  dword ops[5];
  int opnum=0;

  if (x+dx>width) 
    dx=width-x;
  len=dx*bpp;
  if (command==BT848_RISC_WRITEC) {
    dword dx2=BURST_MAX/bpp;
    while (len>BURST_MAX) {
      write_risc_segment(rp, line_adr, command, x, dx2, bpp, width);
      dx-=dx2; x+=dx2;
      len=dx*bpp;
    }
  }
  flags = ((bpp==4) ? BT848_RISC_BYTE3 : 0);
  if (x==0) {
    if (command==BT848_RISC_SKIP) {
      if (dx<width) {
	flags|=BT848_RISC_BYTE_ALL;
	command=BT848_RISC_WRITE;
      }
    }
    if (command==BT848_RISC_WRITEC)
      command=BT848_RISC_WRITE;
    flags|=BT848_RISC_SOL;
  }
  if (x+dx==width)
    flags|=BT848_RISC_EOL;

#ifdef VIRTRISC
  ops[opnum++]=command|flags|len;
  if (command==BT848_RISC_WRITE) {
    ops[opnum++]=line_adr+x*bpp;
  }
  write_risc_virt(rp,opnum,ops);
#else
/*  printk("bttv: %p %d %d\n",line_adr,x,dx); */
  *((*rp)++)=command|flags|len;
  if (command==BT848_RISC_WRITE)
    *((*rp)++)=line_adr+x*bpp;
#endif
}


static void 
make_clip_tab(struct bttv *btv)
{
  int i,ncr;
  uint yy, y, x, dx;
  struct cliprec first, *cur, *cur2, *nx, first2, *prev, *nx2, *cr;
  ushort bpp, bpl, width, height, inter;
  dword **rp,*ro,*re;
  dword adr;

  inter = (btv->win.interlace&1)^1;
  bpp = btv->win.bpp;
  bpl = btv->win.bpl;
  ncr = btv->ncr;
  cr = btv->cliprecs;
  ro = btv->risc_odd;
  re = btv->risc_even;
  width = btv->win.width;
  height = btv->win.height;
  adr = btv->win.vidadr+btv->win.x*bpp+btv->win.y*bpl;

  first.next=NULL;
  for (i=0; i<ncr; i++) {
    cur=&first;
    while ((nx=cur->next) && (cr[i].y > cur->next->y))
      cur=nx; 
    cur->next=&(cr[i]);
    cr[i].next=nx;
  }
  first2.next=NULL;

  *(ro++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1; *(ro++)=0;
  *(re++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1; *(re++)=0;

  /* loop through all lines */
  for (yy=0; yy<(height<<inter); yy++) {
    y=yy>>inter;
    rp= (yy&1) ? &re : &ro;

    /* remove rects with y2 > y */
    if ((cur=first2.next)) {
      prev=&first2;
      do {
	if (cur->y2 < y) 
	  prev->next=cur->next;
	else
	  prev=cur;
      } while ((cur=cur->next));
    }

    /* add rect to second (x-sorted) list if rect.y == y  */
    if ((cur=first.next)) {
      while ((cur) && (cur->y == y)) { 
	first.next=cur->next;
	cur2=&first2;
	while ((nx2=cur2->next) && (cur->x > cur2->next->x)) 
	  cur2=nx2; 
	cur2->next=cur;
	cur->next=nx2;
	cur=first.next;
      }
    }

/*    if (btv->win.y+yy<0) {
      write_risc_segment(rp, adr, BT848_RISC_SKIP, x, dx, bpp, width);
      continue;
    }
*/
    dx=x=0;
    for (cur2=first2.next; cur2; cur2=cur2->next) {
      if (x+dx < cur2->x) {
	if (dx)
	  write_risc_segment(rp, adr, BT848_RISC_SKIP, x, dx, bpp, width);
	x+=dx;
	dx=cur2->x-x;
	write_risc_segment(rp, adr, BT848_RISC_WRITEC, x, dx, bpp, width);
	x=cur2->x;
	dx=cur2->x2-x+1;
      } else
	if (x+dx < cur2->x2) 
	  dx=cur2->x2-x+1;
    }
    if (dx) {
      write_risc_segment(rp, adr,   BT848_RISC_SKIP, x, dx, bpp, width);
      x+=dx;
    }
    if ((dx=width-x))
      write_risc_segment(rp, adr, BT848_RISC_WRITEC, x, dx, bpp, width);
    if ((!inter)||(yy&1))
      adr+=bpl;
  }

  *(ro++) = BT848_RISC_JUMP;
  *(ro++) = btv->bus_vbi_even;
  *(re++) = BT848_RISC_JUMP;
  *(re++) = btv->bus_vbi_odd;
}

struct tvnorm {
  ushort cropwidth, cropheight;
  ushort totalwidth;
  u_char adelay, bdelay, iform;
  ulong xsfnum, xsfden;
  ushort hdelayx1, hactivex1;
  ushort vdelay;
};

static struct tvnorm tvnorms[] = {
  /* PAL-BDGHI */
  { 768, 576, 1135, 0x7f, 0x72, (BT848_IFORM_PAL_BDGHI|BT848_IFORM_XT1),
    944, 768, 186, 922, 0x20},
  /* NTSC */
  { 640, 480,  910, 0x68, 0x5d, (BT848_IFORM_NTSC|BT848_IFORM_XT0),
    780, 640, 135, 754, 0x16},
  /* SECAM */
  { 768, 576, 1135, 0x7f, 0x00, (BT848_IFORM_SECAM|BT848_IFORM_XT1),
    944, 768, 186, 922, 0x20},
  /* PAL-M */
  { 640, 480, 910, 0x68, 0x5d, (BT848_IFORM_PAL_M|BT848_IFORM_XT0),
    780, 640, 186, 922, 0x16},
  /* PAL-N */
  { 768, 576, 1135, 0x7f, 0x72, (BT848_IFORM_PAL_N|BT848_IFORM_XT1),
    944, 768, 186, 922, 0x20},
};
#define TVNORMS (sizeof(tvnorms)/sizeof(tvnorm))

static void
bt848_set_geo(struct bttv *btv, ushort width, ushort height, ushort fmt)
{
  ushort vscale, hscale;
  ulong xsf, sr;
  ushort hdelay, vdelay;
  ushort hactive, vactive;
  ushort inter;
  unchar crop;
  struct tvnorm *tvn;

  if (!width || !height)
    return;
  btwrite(fmt, BT848_COLOR_FMT);

  tvn=&tvnorms[btv->win.norm];
  hactive=width;
  if (hactive < 193) {
    btwrite (2, BT848_E_VTC);
    btwrite (2, BT848_O_VTC);
  } else if (hactive < 385) {
    btwrite (1, BT848_E_VTC);
    btwrite (1, BT848_O_VTC);
  } else {
    btwrite (0, BT848_E_VTC);
    btwrite (0, BT848_O_VTC);
  }
    
  btv->win.cropwidth=tvn->cropwidth;
  btv->win.cropheight=tvn->cropheight;
  inter=(btv->win.interlace&1)^1;
  btwrite(tvn->adelay, BT848_ADELAY);
  btwrite(tvn->bdelay, BT848_BDELAY);
  btaor(tvn->iform,~(BT848_IFORM_NORM|BT848_IFORM_XTBOTH), BT848_IFORM);
  xsf = (width*tvn->xsfnum)/tvn->xsfden;
  hscale = ((tvn->totalwidth*4096UL)/xsf-4096);
  vdelay=btv->win.cropy+tvn->vdelay;
  hdelay=((hactive*tvn->hdelayx1)/tvn->hactivex1+btv->win.cropx)&0x3fe;
  sr=((btv->win.cropheight>>inter)*512)/height-512;
  vscale=(0x10000UL-sr)&0x1fff;
  vactive=btv->win.cropheight;

  if (btv->win.interlace) { 
    btor(BT848_VSCALE_INT, BT848_E_VSCALE_HI);
    btor(BT848_VSCALE_INT, BT848_O_VSCALE_HI);
  } else {
    btand(~BT848_VSCALE_INT, BT848_E_VSCALE_HI);
    btand(~BT848_VSCALE_INT, BT848_O_VSCALE_HI);
  }

  btwrite(hscale>>8, BT848_E_HSCALE_HI);
  btwrite(hscale>>8, BT848_O_HSCALE_HI);
  btwrite(hscale&0xff, BT848_E_HSCALE_LO);
  btwrite(hscale&0xff, BT848_O_HSCALE_LO);

  btwrite((vscale>>8)|(btread(BT848_E_VSCALE_HI)&0xe0), BT848_E_VSCALE_HI);
  btwrite((vscale>>8)|(btread(BT848_O_VSCALE_HI)&0xe0), BT848_O_VSCALE_HI);
  btwrite(vscale&0xff, BT848_E_VSCALE_LO);
  btwrite(vscale&0xff, BT848_O_VSCALE_LO);

  btwrite(hactive&0xff, BT848_E_HACTIVE_LO);
  btwrite(hactive&0xff, BT848_O_HACTIVE_LO);
  btwrite(hdelay&0xff, BT848_E_HDELAY_LO);
  btwrite(hdelay&0xff, BT848_O_HDELAY_LO);

  btwrite(vactive&0xff, BT848_E_VACTIVE_LO);
  btwrite(vactive&0xff, BT848_O_VACTIVE_LO);
  btwrite(vdelay&0xff, BT848_E_VDELAY_LO);
  btwrite(vdelay&0xff, BT848_O_VDELAY_LO);

  crop=((hactive>>8)&0x03)|((hdelay>>6)&0x0c)|
    ((vactive>>4)&0x30)|((vdelay>>2)&0xc0);
  btwrite(crop, BT848_E_CROP);
  btwrite(crop, BT848_O_CROP);
}

static void 
bt848_set_winsize(struct bttv *btv)
{
  ushort norm;

  switch (btv->win.bpp) {
    /* RGB8 seems to be a 9x5x5 GRB color cube starting at color 16
       Why the h... can't they even mention this in the datasheet??? */
  case 1: 
    norm=BT848_COLOR_FMT_RGB8;
    btand(~0x10, BT848_CAP_CTL); /* Dithering looks much better in this mode */
    break;
  case 2: 
    norm=BT848_COLOR_FMT_RGB16;
    btor(0x10, BT848_CAP_CTL);
    break;
  case 3: 
    norm=BT848_COLOR_FMT_RGB24;
    btor(0x10, BT848_CAP_CTL);
    break;
  case 4: 
    norm=BT848_COLOR_FMT_RGB32;
    btor(0x10, BT848_CAP_CTL);
    break;
  }
  bt848_set_geo(btv,btv->win.width, btv->win.height, norm);
}


static struct tunertype tuners[] = {
   {"Temic PAL", TEMIC, PAL,
      16*140.25,16*463.25,0x02,0x04,0x01,0x8e,0xc2,623},
   {"Philips PAL_I", Philips, PAL_I,
      16*140.25,16*463.25,0xa0,0x90,0x30,0x8e,0xc0,623},
   {"Philips NTSC", Philips, NTSC,
      16*157.25,16*451.25,0xA0,0x90,0x30,0x8e,0xc0,732},
   {"Philips SECAM", Philips, SECAM,
      16*168.25,16*447.25,0xA3,0x93,0x33,0x8e,0xc0,623},
   {"NoTuner", NoTuner, NOTUNER,
      0        ,0        ,0x00,0x00,0x00,0x00,0x00,000},
   {"Philips PAL", Philips, PAL,
      16*168.25,16*447.25,0xA0,0x90,0x30,0x8e,0xc0,623},
   {"Temic NTSC", TEMIC, NTSC,
      16*157.25,16*463.25,0x02,0x04,0x01,0x8e,0xc2,732},
   {"TEMIC PAL_I", TEMIC, PAL_I,
      16*170.00,16*450.00,0xa0,0x90,0x30,0x8e,0xc2,623},
}; 

/* set TSA5522 synthesizer frequency in 1/16 Mhz steps*/
static void 
set_freq(struct bttv *btv, ushort freq)
{
  unchar config;
  ushort div;
  struct tunertype *tun=&tuners[btv->tuner];
  int oldAudio = btv->audio;

  audio(btv, AUDIO_MUTE);
  udelay(AUDIO_MUTE_DELAY);
  if (freq < tun->thresh1) 
    config = tun->VHF_L;
  else if (freq < tun->thresh2) 
    config = tun->VHF_H;
  else
    config = tun->UHF;

  div=freq+tun->IFPCoff;
  
  div&=0x7fff;
  if (I2CWrite(btv, btv->tuneradr, (div>>8)&0x7f, div&0xff, 1)<0)
    return;
  I2CWrite(btv, btv->tuneradr, tun->config, config, 1);
  if (!(oldAudio & AUDIO_MUTE))
    {
      udelay(FREQ_CHANGE_DELAY);
      audio(btv, AUDIO_UNMUTE);
    }
}

static void set_ffreq(struct bttv *btv, float freq)
{
  set_freq(btv, freq*16);
}

#ifdef LINUX_2_1
static long 
bttv_read(struct inode *ino, struct file *file, char *buf, unsigned long count)
#else
static int
bttv_read(struct inode *ino, struct file *file, char *buf, int count)
#endif
{
  unsigned int minor = MINOR(ino->i_rdev);
  struct bttv *btv;
  int q,todo;

  btv=&bttvs[minor&15];

  todo=count;
  while (todo>(q=VBIBUF_SIZE-btv->vbip)) {
    copy_to_user((void *) buf, (void *) btv->vbibuf+btv->vbip, q);
    todo-=q;
    buf+=q;

    if (todo) 
      interruptible_sleep_on(&btv->vbiq);
  }
  if (todo) {
    copy_to_user((void *) buf, (void *) btv->vbibuf+btv->vbip, todo);
    btv->vbip+=todo;
  }
  return count;
}

/*
static long long 
bttv_lseek(struct inode * inode, struct file * file,
	    long long offset, int orig)
{
  return -EINVAL;
}
*/

static int
bttv_open( struct inode* ino, struct file* filep)
{
  unsigned int minor = MINOR(ino->i_rdev);
  struct bttv *btv;
  int users, i;

  if ((minor&15) >= bttv_num)
    return -ENODEV;
  btv=&bttvs[minor&15];
 
  switch (minor&0xf0) {
  case 0x00:
    if (btv->user)
      return -EBUSY;
    btv->user++;
    audio(btv, AUDIO_UNMUTE);
    for (i=users=0; i<bttv_num; i++)
      users+=bttvs[i].user;
    if (users==1)
      find_vga();
    break;
  case 0x10:
    break;
  case 0x20:
    btv->vbip=VBIBUF_SIZE;
    btv->cap|=0x0c;
    bt848_set_risc_jmps(btv);
    break;
  }
  MOD_INC_USE_COUNT;
  return 0;   
}

static
#if LINUX_VERSION_CODE >= 0x020100
int
#else
void
#endif
bttv_close( struct inode* ino, struct file* filep)
{
  unsigned int minor = MINOR(ino->i_rdev);
  struct bttv *btv;
  
  if ((minor&15) >= bttv_num)
#if LINUX_VERSION_CODE >= 0x020100
    return -ENODEV;
#else
  return;
#endif
  
  btv=&bttvs[minor&15];

  if (minor<16) { 
    btv->user--;
    audio(btv, AUDIO_MUTE);
    btv->cap&=~3;
  }
  if (minor&0x20) {
    btv->cap&=~0x0c;
  }
  bt848_set_risc_jmps(btv);

  MOD_DEC_USE_COUNT;
  
#if LINUX_VERSION_CODE >= 0x020100
  return 0;
#endif
}

int fmtbppx2[16] = {
  8, 6, 4, 4, 4, 3, 2, 2, 4, 3, 0, 0, 0, 0, 2, 0 
};

static struct gbuffer *
alloc_gbuffer(ushort width, ushort height, uint fmt, uint interlace)
{
  struct gbuffer *gbuf;
  if (!(gbuf=(struct gbuffer *) kmalloc(sizeof(struct gbuffer))))
    return 0;
  
  gbuf->fmt=fmt;
  gbuf->width=width;
  gbuf->height=height;
  gbuf->bpl=(width*fmtbppx2[fmt&0x0f])>>1;
  gbuf->len=height*gbuf->bpl;
  if (!(gbuf->adr=vmalloc(gbuf->len+8192))) {
    kfree(gbuf);
    return 0;
  }
  gbuf->ro=gbuf->adr+gbuf->len;
  gbuf->re=gbuf->ro+4096;
  gbuf->bro=kvirt_to_bus((ulong)gbuf->ro);
  gbuf->bre=kvirt_to_bus((ulong)gbuf->re);
}

/*
static int
gbuffer_maketab(struct gbuffer *gbuf)
{
  ulong i;
  ulong bpl; 
  ulong bl;
  dword *ro, *re, *rp;
  dword adr, busadr;

  ro=gbuf->ro;
  re=gbuf->re;
  bpl=gbuf->bpl;
  adr=gbuf->adr;
  busadr=kvirt_to_bus(adr);
  
  *(ro++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1|(14<<20); *(ro++)=0;
  *(re++)=BT848_RISC_SYNC|BT848_FIFO_STATUS_FM1; *(re++)=0;
  
  for (i=0; i < gbuf->height; i++) {
    if (gbuf->interlace) 
      rp= (i&1) ? &re : &ro;
    else 
      rp= (i>gbuf->height) ? &re : &ro; 

    if (bpl<=(PAGE_SIZE) {
      *((*rp)++)=BT848_RISC_WRITE|BT848_RISC_SOL|BT848_RISC_EOL|bpl; 
      *((*rp)++)=busadr;
      busadr+=bpl; 
      if (!iovl) {
	iov++;
	iovl=iov->iov_len;
	adr=kvirt_to_bus(gbuf->adr+);
      }
    } else {
      *((*rp)++)=BT848_RISC_WRITE|BT848_RISC_SOL|iovl;
      *((*rp)++)=iovp;
      bl=bpl-iovl;
      iov++;
      while (bl>iov->iov_len) {
	*((*rp)++)=BT848_RISC_WRITE|iov->iov_len;
	*((*rp)++)=(dword) iov->iov_base;
	bl-=iov->iov_len;
	iov++;
      }
      *((*rp)++)=BT848_RISC_WRITE|BT848_RISC_EOL|bl;
      *((*rp)++)=(dword) iov->iov_base;
      iovl=iov->iov_len-bl; iovp=bl+(dword)iov->iov_base;      
    }
  }

  *(ro++)=BT848_RISC_JUMP;
  *(ro++)=btv->bus_vbi_even;
  *(re++)=BT848_RISC_JUMP|BT848_RISC_IRQ|(2<<16);
  *(re++)=btv->bus_vbi_odd;

  return 0;
}
*/


char Ximg_locked,Ximg_ready;
XImage X_img;
struct iovec *X_iovecs;
int X_iovcnt;
dword *X_ro, *X_re;

static char Lock_Ximage(struct bttv *btv, XImage *ximg)
{
  int i,j;

  if (Ximg_locked == 1)
     {
#ifdef DEBUG
      printk("bttv: Ximg already locked\n");
#endif     
      return 1;
     } 

  if (ximg == NULL)
     return -1;

  X_img.width       	 = ximg->width;
      
#ifdef DEBUG
  printk("ximg->width %d, ximg->height %d,\n",ximg->width,ximg->height);    
#endif

  X_img.height      	 = ximg->height;
  X_img.xoffset     	 = ximg->xoffset;
  X_img.format      	 = ximg->format;
  X_img.data        	 = ximg->data;
  X_img.byte_order  	 = ximg->byte_order;
  X_img.bitmap_unit      = ximg->bitmap_unit; 
  X_img.bitmap_bit_order = ximg->bitmap_bit_order;
  X_img.bitmap_pad 	 = ximg->bitmap_pad;
  X_img.depth 		 = ximg->depth;
  X_img.bytes_per_line   = ximg->bytes_per_line;
  X_img.bits_per_pixel   = ximg->bits_per_pixel;
  X_img.red_mask 	 = ximg->red_mask;
  X_img.green_mask       = ximg->green_mask;
  X_img.blue_mask 	 = ximg->blue_mask;

  if (
      (X_img.width > 384) ||
      (X_img.height > 288) ||
      (X_img.xoffset > 0) ||
      (X_img.bits_per_pixel != X_img.depth) ||
      (X_img.format != ZPixmap) ||
      (X_img.bytes_per_line != (X_img.width * (X_img.depth / 8))) ||
     ((X_img.depth != 8) && (X_img.depth != 16) && (X_img.depth != 24))
     )
     {
#ifdef DEBUG     
      printk("bttv: invalid XImage supplied\n");
#endif      
      return -1;
     } 

  if (!(X_iovecs = vmalloc(sizeof(struct iovec)*1024)))
     {
#ifdef DEBUG
      printk("bttv: vmalloc X_iovecs error\n");
#endif     
      return -2;
     } 

#ifdef DEBUG
  printk("locking mem @ 0x%08lx with size %ld from XImage\n",
          (long)X_img.data,X_img.width*X_img.height*X_img.depth/8);
#endif
  X_iovcnt = lock_user((unsigned long)X_img.data, BTTV_LOCKMEMSIZE, 1024, X_iovecs);

  if (X_iovcnt < 0) 
     {
      printk("bttv: X_iovcnt\n");
      return -4;
     }
/*
   for (i=0; i<10; i++)
   printk("bttv: 0x%08x 0x%08x\n", iovecs[i].iov_base, iovecs[i].iov_len);
   for (i=j-10; i<j; i++)
   printk("bttv: 0x%08x 0x%08x\n", iovecs[i].iov_base, iovecs[i].iov_len);
 */

  if (!(X_ro = (dword *) kmalloc(4072, GFP_KERNEL))) 
     {
      vfree(X_iovecs);
      printk("bttv: kmalloc X_ro\n");
      return -5;
     }

  if (!(X_re = (dword *) kmalloc(4072, GFP_KERNEL))) 
     {
      kfree((void *) X_ro);
      vfree(X_iovecs);
      printk("bttv: kmalloc X_re\n");
      return -6;
     }

  /* success */
#ifdef DEBUG
  printk("locking ximage succeed\n"); 
#endif  
  return 0;
}  

void Unlock_Ximage(struct bttv *btv)
{
  bt848_dma(btv,0);
  bt848_cap(btv, 0);         /* capturing off */

  unlock_user(X_iovcnt, X_iovecs);
  
  kfree((void *) X_ro);
  kfree((void *) X_re);
  vfree(X_iovecs);
#ifdef DEBUG
  printk("unlock ximage succeed\n");
#endif  
}

static char Fill_Ximage(struct bttv *btv)
{
  int i,j,k,m,n,o,p,q;
/*  dword *ro, *re;*/
/*  XImage *ximg; */

  if (Ximg_locked != 1)
     {
#ifdef DEBUG
      printk("bttv: Ximg not locked\n");
#endif          
      return -1;
     } 

  btv->win.bpp = X_img.depth / 8;
  btv->win.width = X_img.width;
  btv->win.height = X_img.height;
  btv->win.bpl = X_img.bytes_per_line; 
 
#ifdef DEBUG
  printk("grabbing direct into ximage @ 0x%08lx\n",(long)X_img.data); 
#endif
  
  bt848_set_winsize(btv);
  
  bt848_dma(btv,0);
  bt848_cap(btv, 0);

  make_iov_risctab(btv, X_ro, X_re, X_iovcnt, X_iovecs);

  btv->risc_jmp[5] = virt_to_bus(X_ro);
  btv->risc_jmp[11] = virt_to_bus(X_re);

  btor(3, BT848_CAP_CTL);
  btor(3, BT848_GPIO_DMA_CTL);

  /* wait for wakeup caused by RISCI *
  interruptible_sleep_on(&btv->capq);
   */
   
  Ximg_ready = 0; 

#ifdef DEBUG
  printk("bttv: XFill succeed @ 0x%08x\n", (long)X_img.data); 
#endif  
  
  return 0;
}


static void 
testm(struct bttv *btv, unsigned long mem)
{
  struct iovec *iovecs;
  int iovcnt;
  int i,j;
  dword *ro, *re;

  if (!(iovecs = vmalloc(sizeof(struct iovec)*1024)))
    return;
  /*printk("bttv: mem: 0x%08x\n", (int) mem);*/
  iovcnt=lock_user(mem, BTTV_LOCKMEMSIZE, 1024, iovecs);
  /*printk("bttv: iovnum: %d\n", iovcnt);*/
  if (iovcnt<0) {
    printk("bttv: testm_error: %d\n", j);
    return;
  }
/*
   for (i=0; i<10; i++)
   printk("bttv: 0x%08x 0x%08x\n", iovecs[i].iov_base, iovecs[i].iov_len);
   for (i=j-10; i<j; i++)
   printk("bttv: 0x%08x 0x%08x\n", iovecs[i].iov_base, iovecs[i].iov_len);
*/

  if (!(ro=(dword *) kmalloc(4072, GFP_KERNEL))) {
    vfree(iovecs);
    return;
  }
  if (!(re=(dword *) kmalloc(4072, GFP_KERNEL))) {
    kfree((void *) ro);
    vfree(iovecs);
    return;
  }

  bt848_dma(btv,0);
  bt848_cap(btv, 0);
  make_iov_risctab(btv, ro, re, iovcnt, iovecs);
  btv->risc_jmp[5]=virt_to_bus(ro);
  btv->risc_jmp[11]=virt_to_bus(re);
  btor(3, BT848_CAP_CTL);
  btor(3, BT848_GPIO_DMA_CTL);

  /* wait for wakeup caused by RISCI */
  interruptible_sleep_on(&btv->capq);

  unlock_user(iovcnt, iovecs);
  kfree((void *) ro);
  kfree((void *) re);
  vfree(iovecs);
  
/*  printk("bttv: testm succeed\n");*/
  printk("bttv: testm succeed @ 0x%08x\n", (int)mem); 
}


/***********************************/
/* ioctls and supporting functions */
/***********************************/

static inline void
bt848_bright(struct bttv *btv, uint bright)
{
  btwrite(bright&0xff, BT848_BRIGHT);
}

static inline void
bt848_hue(struct bttv *btv, uint hue)
{
  btwrite(hue&0xff, BT848_HUE);
}

static inline void
bt848_contrast(struct bttv *btv, uint cont)
{
  uint conthi;

  conthi=(cont>>6)&4;
  btwrite(cont&0xff, BT848_CONTRAST_LO);
  btaor(conthi, ~4, BT848_E_CONTROL);
  btaor(conthi, ~4, BT848_O_CONTROL);
}

static inline void
bt848_sat_u(struct bttv *btv, ulong data)
{
  ulong datahi;

  datahi=(data>>7)&2;
  btwrite(data&0xff, BT848_SAT_U_LO);
  btaor(datahi, ~2, BT848_E_CONTROL);
  btaor(datahi, ~2, BT848_O_CONTROL);
}

static inline void
bt848_sat_v(struct bttv *btv, ulong data)
{
  ulong datahi;

  datahi=(data>>8)&1;
  btwrite(data&0xff, BT848_SAT_V_LO);
  btaor(datahi, ~1, BT848_E_CONTROL);
  btaor(datahi, ~1, BT848_O_CONTROL);
}

static int
bttv_ioctl(struct inode* node,struct file* file,
	    unsigned int cmd, unsigned long arg)
{
  unchar eedata[256];
  unsigned int minor = MINOR(node->i_rdev);
  unsigned long data;
  int ret;
  float dataf;
  struct bttv *btv=&bttvs[minor];
  static ushort creg;
  XImage ximg;
  
  switch (cmd)
    {
    case BTTV_SET_BTREG:
      copy_from_user((void *) &creg, (void *) arg, sizeof(creg));
      break;
    case BTTV_WRITE_BTREG:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      btwrite(data,creg);
      break;
    case BTTV_READ_BTREG:
      data=btread(creg);
      copy_to_user((void *) arg, (void *) &data, sizeof(data));
      break;
    case BTTV_SETWTW:
      copy_from_user((void *) &btv->win, (void *) arg,
		     sizeof(btv->win));
      bt848_set_winsize(btv);
      break;
    case BTTV_GETWTW:
      copy_to_user((void *) arg, (void *) &btv->win,
		   sizeof(btv->win));
      break;
    case BTTV_DMA:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      btwrite(data, BT848_GPIO_DMA_CTL);
      break;

    case BTTV_CAP_ON:
      bt848_cap(btv, 1);
      break;
    case BTTV_CAP_OFF:
      bt848_cap(btv, 0);
      break;

    case BTTV_SETCLIP:
      if (arg) {
	copy_from_user((void *) btv->cliprecs, (void *) arg,
		       sizeof(struct cliprec)*MAX_CLIPRECS);
	btv->ncr=btv->cliprecs[MAX_CLIPRECS-1].x;
      } else 
	btv->ncr=0;
      make_clip_tab(btv);
      break;
    case BTTV_SETVIRTADR:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      btv->win.vidadr=uvirt_to_bus(data);
      break;
    case BTTV_GETBTTV:
      copy_to_user((void *) arg, (void *) btv, sizeof(struct bttv));
      break;

    case BTTV_SETFREQ:
      copy_from_user((void *) &btv->win.freq,
		     (void *) arg, sizeof(btv->win.freq));
      set_freq(btv, btv->win.freq);
      break;

    case BTTV_SET_FFREQ:
      copy_from_user((void *)&dataf, (void *) arg, sizeof(dataf));
      set_ffreq (btv, dataf);
      break;

    case BTTV_MUTE:
      copy_from_user((void *)&data, (void *)arg, sizeof(data));
      if (data)
	audio(btv, AUDIO_MUTE);
      else
	audio(btv, AUDIO_UNMUTE);
      break;

    case BTTV_INPUT:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      bt848_muxsel(btv, data);
      break;
    case BTTV_WRITEEE:
      copy_from_user((void *) eedata, (void *) arg, 256);
      writeee(btv, eedata);
      break;
    case BTTV_READEE:
      readee(btv, eedata);
      copy_to_user((void *) arg, (void *) eedata, 256);
      break;

    case BTTV_BRIGHT:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      bt848_bright(btv, data);
      break;
    case BTTV_COLOR:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      bt848_sat_u(btv, data); 
      bt848_sat_v(btv, (data*201)/237); 
      break;
    case BTTV_HUE:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      bt848_hue(btv, data);
      break;
    case BTTV_CONTRAST:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      bt848_contrast(btv, data);
      break;
    case BTTV_READ_MSP_REG:
      if (!btv->msp3400c)
	return -EINVAL; /* this one? */
      copy_from_user((void *)&data, (void *) arg, sizeof(data));
      data=read_msp3400c(btv, I2C_MSP3400C_DFP,
			 (data >> 16) & 0xff);
      copy_to_user((void *) arg, (void *) &data, sizeof(data));
      break;
    case BTTV_WRITE_MSP_REG:
      if (!btv->msp3400c)
	return -EINVAL; /* this one? */
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
      write_msp3400c(btv, I2C_MSP3400C_DFP,
		     (data >> 16) & 0xffff, data & 0xffff);
      break;
    case BTTV_TESTM:
      copy_from_user((void *) &data, (void *) arg, sizeof(data));
/*      printk("bttv: testm: 0x%08x\n", (int) data); */
      testm(btv, data);
      break;
    
    case BTTV_LOCK_XIMG:
      copy_from_user((void *) &ximg, (void *) arg, sizeof(ximg));
      ret = Lock_Ximage(btv,&ximg);
      if (ret == 0)
          Ximg_locked = 1;
      return ret;    
      break;

    case BTTV_FILL_XIMG:
      if (Ximg_locked == 1)
         return Fill_Ximage(btv);
      
      return -1;   
      break;
      
    case BTTV_XIMG_READY:
      return Ximg_ready;
      break;  

    case BTTV_UNLOCK_XIMG:
      if (Ximg_locked == 1)
         {
          Unlock_Ximage(btv);
          Ximg_locked = 0;
         } 
      break;

    default:
      return -EINVAL;
    }
  return 0;
}

static struct file_operations bttv_fops = {
	NULL,
	bttv_read,
	NULL,
	NULL,
	NULL,		/* bttv_select */
	bttv_ioctl,
	NULL,           /* bttv_mmap */
	bttv_open,
	bttv_close,
	NULL		/* fsync */
};

struct vidbases {
  ushort vendor, device;
  char *name;
  uint badr;
};

static struct vidbases vbs[] = {
  { PCI_VENDOR_ID_TSENG, 0, 
        "TSENG", PCI_BASE_ADDRESS_0},
        
  { PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_MIL, 
        "Matrox Millennium", PCI_BASE_ADDRESS_1},
        
  { PCI_VENDOR_ID_MATROX, 0x051a, 
        "Matrox Mystique", PCI_BASE_ADDRESS_1},
        
  { PCI_VENDOR_ID_S3, 0, 
        "S3", PCI_BASE_ADDRESS_0},
        
  { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_ATI_210888GX, 
        "ATI MACH64 Winturbo", PCI_BASE_ADDRESS_0},
        
  { PCI_VENDOR_ID_CIRRUS, 0, 
        "Cirrus Logic", PCI_BASE_ADDRESS_0},
        
  { PCI_VENDOR_ID_N9, PCI_DEVICE_ID_N9_I128, 
        "Number Nine Imagine 128", PCI_BASE_ADDRESS_0},
        
  { PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TGA, 
        "DEC DC21030", PCI_BASE_ADDRESS_0},
};


/* DEC TGA offsets stolen from XFree-3.2 */
static uint dec_offsets[4] = {0x200000, 0x804000, 0, 0x1004000};

#define NR_CARDS (sizeof(vbs)/sizeof(struct vidbases))

/* Scan for PCI display adapter
   if more than one card is present the last one is used for now */

static int
find_vga(void)
{
  unsigned int devfn, class, vendev;
  ushort vendor, device, badr;
  int found=0, bus=0, i, tga_type;
  unsigned int vidadr=0;


  for (devfn = 0; devfn < 0xff; devfn++) {
    if (PCI_FUNC(devfn) != 0)
      continue;
    pcibios_read_config_dword(bus, devfn, PCI_VENDOR_ID, &vendev);
    if (vendev == 0xffffffff || vendev == 0x00000000) 
      continue;
    pcibios_read_config_word(bus, devfn, PCI_VENDOR_ID, &vendor);
    pcibios_read_config_word(bus, devfn, PCI_DEVICE_ID, &device);
    pcibios_read_config_dword(bus, devfn, PCI_CLASS_REVISION, &class);
    class = class >> 16;
/*    if (class == PCI_CLASS_DISPLAY_VGA) {*/
    if ((class>>8) == PCI_BASE_CLASS_DISPLAY ||
/* Number 9 GXE64Pro needs this */
	class == PCI_CLASS_NOT_DEFINED_VGA) {
      badr=0;
      printk("bttv: PCI display adapter: ");

      for (i=0; i<NR_CARDS; i++) {
	if (vendor==vbs[i].vendor) {
	  if (vbs[i].device) 
	    if (vbs[i].device!=device)
	      continue;
	  printk("%s.\n", vbs[i].name);
	  badr=vbs[i].badr;
	  break;
	}
      }
      if (!badr) {
	printk(" unknown video mem base %lx address.\n",badr);
	continue;
      }
      pcibios_read_config_dword(bus, devfn, badr, &vidadr);
      if (vidadr & PCI_BASE_ADDRESS_SPACE_IO) {
	printk("bttv: memory seems to be I/O memory.\n");
	printk("bttv: check entry for your card type in bttv.c vidbases struct!\n");
	continue;
      }
      vidadr &= PCI_BASE_ADDRESS_MEM_MASK;
      if (!vidadr) {
	printk("bttv: memory @ 0, must be something wrong!");
	continue;
      }
      
      if (vendor==PCI_VENDOR_ID_DEC)
	if (device==PCI_DEVICE_ID_DEC_TGA) {
	  tga_type = (readl((unsigned long)vidadr) >> 12) & 0x0f;
 	  if (tga_type != 0 && tga_type != 1 && tga_type != 3) {
	    printk("TGA type (0x%x) unrecognized!\n", tga_type);
	    found--;
 	  }
 	  vidadr+=dec_offsets[tga_type];
	}

      printk("bttv: memory @ 0x%08x, ", vidadr);
      printk("devfn: 0x%04x.\n", devfn);
      
      found++;
    }
  }
  
  if (vidmem) {
    vidadr=vidmem<<20;
#ifdef DEBUG    
    printk("bttv: videomem override: 0x%08x\n", vidadr);
#endif    
    found=1;
  }
  for (i=0; i<BTTV_MAX; i++)
    bttvs[i].win.vidadr=vidadr;

  return found;
}

#define  TRITON_PCON	           0x50 
#define  TRITON_BUS_CONCURRENCY   (1<<0)
#define  TRITON_STREAMING	  (1<<1)
#define  TRITON_WRITE_BURST	  (1<<2)
#define  TRITON_PEER_CONCURRENCY  (1<<3)
  
static void
handle_chipset(void)
{
  int index;
  
  /* just in case some nut set this to something dangerous */
  if (triton1)
    triton1=BT848_INT_ETBF;

  for (index = 0; index < 8; index++) {
    unsigned char bus, devfn;
    unchar b, bo;
    
    /* nothing wrong with this one, just checking buffer control config */
    if (!pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82441,
			    index, &bus, &devfn))  {
      pcibios_read_config_byte(bus, devfn, 0x53, &b);
      printk("bttv: Host bridge: 82441FX Natoma, ");
      printk("bufcon=0x%02x\n",b);
    }

    if (!pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82437,
			    index, &bus, &devfn)) {
      printk("bttv: Host bridge 82437FX Triton PIIX\n");
      triton1=BT848_INT_ETBF;

      /* 430FX (Triton I) freezes with bus concurrency on -> switch it off */
/*
      pcibios_read_config_byte(bus, devfn, TRITON_PCON, &b);
      bo=b;
      printk("bttv: 82437FX: PCON: 0x%x\n",b);
      if(!(b & TRITON_BUS_CONCURRENCY)) {
	printk("bttv: 82437FX: disabling bus concurrency\n");
	b |= TRITON_BUS_CONCURRENCY;
      }

      if(b & TRITON_PEER_CONCURRENCY) {
	printk("bttv: 82437FX: disabling peer concurrency\n");
	b &= ~TRITON_PEER_CONCURRENCY;
      }

      if(!(b & TRITON_STREAMING)) {
	printk("bttv: 82437FX: disabling streaming\n");
	b |=  TRITON_STREAMING;
      }

      if (b!=bo) {
	pcibios_write_config_byte(bus, devfn, TRITON_PCON, b); 
	printk("bttv: 82437FX: PCON changed to: 0x%x\n",b);
      }
*/
    }
  }
}

static void
init_tda9850(struct bttv *btv)
{
  I2CWrite(btv, I2C_TDA9850, TDA9850_CON3, 0, 1);
}

/* ----------------------------------------------------------------------- */
/* functions for talking to the MSP3400C Sound processor                   */

void
turn_off_msp3400c(struct bttv *btv)
{
  I2CStart(btv);
  I2CSendByte(btv, I2C_MSP3400C,2000);
  I2CSendByte(btv, 0x00,0);
  I2CSendByte(btv, 0x80,0);
  I2CSendByte(btv, 0x00,0);
  I2CStop(btv);
  udelay(20000);
}

void
turn_on_msp3400c(struct bttv *btv)
{
  I2CStart(btv);
  I2CSendByte(btv, I2C_MSP3400C,2000);
  I2CSendByte(btv, 0x00,0);
  I2CSendByte(btv, 0x00,0);
  I2CSendByte(btv, 0x00,0);
  I2CStop(btv);
  udelay(20000);
}

int
read_msp3400c(struct bttv *btv,int dev, int addr)
{
  int val = 0;
  I2CStart(btv);
  I2CSendByte(btv, I2C_MSP3400C,2000);
  I2CSendByte(btv, dev+1,       0);
  I2CSendByte(btv, addr >> 8,   0);
  I2CSendByte(btv, addr & 0xff, 0);
  I2CStart(btv);
  I2CSendByte(btv, I2C_MSP3400C+1,2000);
  val |= (int)I2CReadByte(btv,0) << 8;
  val |= (int)I2CReadByte(btv,1);
  I2CStop(btv);
  return val;
}

void
write_msp3400c(struct bttv *btv,int dev, int addr, int val)
{
  I2CStart(btv);
  I2CSendByte(btv, I2C_MSP3400C,2000);
  I2CSendByte(btv, dev,         0);
  I2CSendByte(btv, addr >> 8,   0);
  I2CSendByte(btv, addr & 0xff, 0);
  I2CSendByte(btv, val >> 8,    0);
  I2CSendByte(btv, val & 0xff,  0);
  I2CStop(btv);
}

#if 0
void
stat_msp3400c(struct bttv *btv)
{
  int i;
  
  for (i =0x00; i < 0x26; i++) {
    printk("msp3400c dfp 0x%02x: %04x\n",i,
	   read_msp3400c(btv,I2C_MSP3400C_DFP,i));
  }
}
#endif


static struct MSP_INIT_DATA_DEM {
    int fir[6];
    int cdo1;
    int cdo2;
    int dfp_1;
    int dfp_2;
} msp_init_data[] = {
    /* de antenne mono */
    { { 3, 18, 27, 48, 66, 72 }, 0x4c638e, 0x4fc0aa, 0x0030, 0x3000 },
#if 0
    /* de antenne stereo */
    { { 3, 18, 27, 48, 66, 72 }, 0x4c638e, 0x4fc0aa, 0x0020, 0x3001 },
    /* de antenne lang1 */
    { { 3, 18, 27, 48, 66, 72 }, 0x4c638e, 0x4fc0aa, 0x0000, 0x3000 },
    /* de antenne lang2 */
    { { 3, 18, 27, 48, 66, 72 }, 0x4c638e, 0x4fc0aa, 0x0010, 0x3000 },
#endif
};

static void
init_msp3400c(struct bttv *btv, int type)
{
    int i;

    printk("bttv: msp3400c init...\n");
    turn_off_msp3400c(btv);
    turn_on_msp3400c(btv);

    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x00bb, 0x00d0); /* av_dc */

    for (i = 5; i >= 0; i--)                              /* fir 1 */
	write_msp3400c(btv,I2C_MSP3400C_DEM, 0x0001,
		       msp_init_data[type].fir[i]);
    
    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x0005, 0x0004); /* fir 2 */
    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x0005, 0x0040);
    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x0005, 0x0000);
    for (i = 5; i >= 0; i--)
	write_msp3400c(btv,I2C_MSP3400C_DEM, 0x0005,
		       msp_init_data[type].fir[i]);

    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x0083, 0x0480); /* MODE_REG */

    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x0093,
		   msp_init_data[type].cdo1 & 0xfff);
    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x009b,
		   msp_init_data[type].cdo1 >> 12);
    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x00a3,
		   msp_init_data[type].cdo2 & 0xfff);
    write_msp3400c(btv,I2C_MSP3400C_DEM, 0x00ab,
		   msp_init_data[type].cdo2 >> 12);

    
    write_msp3400c(btv,I2C_MSP3400C_DFP, 0x0000, 0x7300); /* loudspeaker on */

    write_msp3400c(btv,I2C_MSP3400C_DFP, 0x0008,
		   msp_init_data[type].dfp_1);
    write_msp3400c(btv,I2C_MSP3400C_DFP, 0x000e,
		   msp_init_data[type].dfp_2);

#if 0
    write_msp3400c(btv,I2C_MSP3400C_DFP, 0x0014, 0x1040); /* debug beep */
#endif
    udelay(20000);
}

/* ------------------------------------------------------------------------ */

/* Figure out card and tuner type */
static void
idcard(struct bttv *btv)
{
  int i;

  btwrite(0, BT848_GPIO_OUT_EN);
  printk("bttv: GPIO: 0x%08x\n", btread(BT848_GPIO_DATA));

  btv->type=BTTV_MIRO;
  btv->tuner=tuner;
  
  if (I2CRead(btv, I2C_HAUPEE)>=0)
    btv->type=BTTV_HAUPPAUGE;
  else
    if (I2CRead(btv, I2C_STBEE)>=0)
      btv->type=BTTV_STB;

  for (i=0xc0; i<0xd0; i+=2)
    if (I2CRead(btv, i)>=0) {
      btv->tuneradr=i;
      break;
    }

  btv->dbx      = I2CRead(btv, I2C_TDA9850)  ? 0 : 1;
  btv->msp3400c = I2CRead(btv, I2C_MSP3400C) ? 0 : 1;
  
  if (btv->dbx)
    init_tda9850(btv);
  if (btv->msp3400c)
    init_msp3400c(btv,0);

  /* How do I detect the tuner type for other cards but Miro ???
     Probably somehow encoded into the EEPROM data!?! */
  printk("bttv: model: ");
  switch (btv->type) {
  case BTTV_MIRO:
    btv->tuner=((btread(BT848_GPIO_DATA)>>10)-1)&7;
    printk("MIRO");
    break;
  case BTTV_HAUPPAUGE:
    printk("HAUPPAUGE");
    break;
  case BTTV_STB: 
    printk("STB");
    break;
  case BTTV_INTEL: 
    printk("Intel");
    break;
  case BTTV_DIAMOND: 
    printk("Diamond");
    break;
  }
  printk(" (%s @ 0x%02x)\n", tuners[btv->tuner].name, btv->tuneradr);
  audio(btv, AUDIO_MUTE);
}


static void
bt848_set_risc_jmps(struct bttv *btv)
{
  int flags=btv->cap;
  
  btv->risc_jmp[0]=BT848_RISC_SYNC|BT848_RISC_RESYNC|BT848_FIFO_STATUS_VRE;
  btv->risc_jmp[1]=0;

  btv->risc_jmp[2]=BT848_RISC_JUMP;
  if (flags&8)
    btv->risc_jmp[3]=virt_to_bus(btv->vbi_odd);
  else
    btv->risc_jmp[3]=virt_to_bus(btv->risc_jmp+4);

  btv->risc_jmp[4]=BT848_RISC_JUMP;
  if (flags&2)
    btv->risc_jmp[5]=virt_to_bus(btv->risc_odd);
  else
    btv->risc_jmp[5]=virt_to_bus(btv->risc_jmp+6);

  btv->risc_jmp[6]=BT848_RISC_SYNC|BT848_RISC_RESYNC|BT848_FIFO_STATUS_VRO;
  btv->risc_jmp[7]=0;

  btv->risc_jmp[8]=BT848_RISC_JUMP;
  if (flags&4)
    btv->risc_jmp[9]=virt_to_bus(btv->vbi_even);
  else
    btv->risc_jmp[9]=virt_to_bus(btv->risc_jmp+10);

  btv->risc_jmp[10]=BT848_RISC_JUMP;
  if (flags&1)
    btv->risc_jmp[11]=virt_to_bus(btv->risc_even);
  else
    btv->risc_jmp[11]=virt_to_bus(btv->risc_jmp);

  btaor(flags, ~0x0f, BT848_CAP_CTL);
  if (flags&0x0f)
    bt848_dma(btv, 3);
  else
    bt848_dma(btv, 0);
}


static int
init_bt848(struct bttv *btv)
{
   /* reset the bt848 */
  btwrite(0,BT848_SRESET);

  btv->user=0; 

  printk("bttv: bt848_mem: 0x%p\n",btv->bt848_mem);

  /* default setup for max. PAL size in a 1024xXXX hicolor framebuffer */

  btv->win.norm=0; /* change this to 1 for NTSC, 2 for SECAM */
  btv->win.interlace=1;
  btv->win.x=0;
  btv->win.y=0;
  btv->win.width=768; /* 640 */
  btv->win.height=576; /* 480 */
  btv->win.cropx=0;
  btv->win.cropy=0;
  btv->win.bpp=3;
  btv->win.bpl=1024*btv->win.bpp;
  btv->win.swidth=1024;
  btv->win.sheight=768;
  btv->cap=0;
  
  btv->ncr=0;
  btv->cliprecs=NULL;
  
  
  if (!(btv->risc_odd=(dword *) kmalloc(RISCMEM_LEN/2, GFP_KERNEL)))
    return -1;
  if (!(btv->risc_even=(dword *) kmalloc(RISCMEM_LEN/2, GFP_KERNEL)))
    return -1;

  if (!(btv->risc_jmp =(dword *) kmalloc(1024, GFP_KERNEL)))
    return -1;
  btv->vbi_odd=btv->risc_jmp+12;
  btv->vbi_even=btv->vbi_odd+256;
  btv->bus_vbi_odd=virt_to_bus(btv->risc_jmp);
  btv->bus_vbi_even=virt_to_bus(btv->risc_jmp+6);

  btwrite(virt_to_bus(btv->risc_jmp+2), BT848_RISC_STRT_ADD);

  if (!(btv->vbibuf=(unchar *) vmalloc(VBIBUF_SIZE)))
      return -1;
  if (!(btv->cliprecs=
	(struct cliprec *) vmalloc(sizeof(struct cliprec)*MAX_CLIPRECS)))
      return -1;

  
      

  bt848_muxsel(btv, 1);
  bt848_set_winsize(btv);

/*  btwrite(0, BT848_TDEC); */
  btwrite(0x10, BT848_COLOR_CTL);

  btwrite(0x00, BT848_CAP_CTL);

  btwrite(0x0ff, BT848_VBI_PACK_SIZE);
  btwrite(1, BT848_VBI_PACK_DEL);

  btwrite(0xfc, BT848_GPIO_DMA_CTL);
  btwrite(BT848_IFORM_MUX1 | BT848_IFORM_XTAUTO | BT848_IFORM_PAL_BDGHI,
	  BT848_IFORM);

  bt848_bright(btv, 0x10);
  btwrite(0xd8, BT848_CONTRAST_LO);

  btwrite(0x60, BT848_E_VSCALE_HI);
  btwrite(0x60, BT848_O_VSCALE_HI);
  btwrite(/*BT848_ADC_SYNC_T|*/
	  BT848_ADC_RESERVED|BT848_ADC_CRUSH, BT848_ADC);

  btwrite(BT848_CONTROL_LDEC, BT848_E_CONTROL);
  btwrite(BT848_CONTROL_LDEC, BT848_O_CONTROL);
  btwrite(0x00, BT848_E_SCLOOP);
  btwrite(0x00, BT848_O_SCLOOP);

  btwrite(0xffffffUL,BT848_INT_STAT);
/*	  BT848_INT_PABORT|BT848_INT_RIPERR|BT848_INT_PPERR|BT848_INT_FDSR|
	  BT848_INT_FTRGT|BT848_INT_FBUS|*/
  btwrite(triton1|
	  BT848_INT_SCERR|
	  BT848_INT_RISCI|BT848_INT_OCERR|BT848_INT_VPRES|
	  BT848_INT_FMTCHG|BT848_INT_HLOCK,
	  BT848_INT_MASK);

  make_vbitab(btv);
  bt848_set_risc_jmps(btv);
 
  idcard(btv);

  Ximg_locked = 0;
  Ximg_ready = 0;
  
  return 0;
}

#undef IRQDEBUG
static void 
bttv_irq(int irq, void *dev_id, struct pt_regs * regs)
{
  ulong stat,astat;
  ulong dstat;
  int count;
  struct bttv *btv;
 
  btv=(struct bttv *)dev_id;
  count=0;
  while (1) {
    /* get/clear interrupt status bits */
    stat=btread(BT848_INT_STAT);
    astat=stat&btread(BT848_INT_MASK);
    if (!astat)
      return;
    btwrite(astat,BT848_INT_STAT);
    /* get device status bits */
    dstat=btread(BT848_DSTATUS);
    
    if (astat&BT848_INT_FMTCHG) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_FMTCHG\n");
#endif
      /* btv->win.norm&=(dstat&BT848_DSTATUS_NUML) ? (~1) : (~0);*/
    }
    if (astat&BT848_INT_VPRES) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_VPRES\n");
#endif
    }
    if (astat&BT848_INT_VSYNC) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_VSYNC\n");
#endif
    }
    if (astat&BT848_INT_SCERR) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_SCERR\n");
#endif
      bt848_dma(btv, 0);
      bt848_dma(btv, 1);
      wake_up_interruptible(&btv->vbiq);
      wake_up_interruptible(&btv->capq);
      Ximg_ready = 1;
    }
    if (astat&BT848_INT_RISCI) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_RISCI\n");
#endif
/*    printk ("bttv: IRQ_RISCI%d\n",stat>>28);*/
      if (stat&(1<<28)) {
	btv->vbip=0;
	wake_up_interruptible(&btv->vbiq);
        Ximg_ready = 1;
      }
      if (stat&(2<<28)) {
	bt848_set_risc_jmps(btv);
	wake_up_interruptible(&btv->capq);
        Ximg_ready = 1;
	break;
      }
    }
    if (astat&BT848_INT_OCERR) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_OCERR\n");
#endif
    }
    if (astat&BT848_INT_PABORT) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_PABORT\n");
#endif
    }
    if (astat&BT848_INT_RIPERR) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_RIPERR\n");
#endif
    }
    if (astat&BT848_INT_PPERR) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_PPERR\n");
#endif
    }
    if (astat&BT848_INT_FDSR) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_FDSR\n");
#endif
    }
    if (astat&BT848_INT_FTRGT) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_FTRGT\n");
#endif
    }
    if (astat&BT848_INT_FBUS) {
#ifdef IRQDEBUG
      printk ("bttv: IRQ_FBUS\n");
#endif
    }
    if (astat&BT848_INT_HLOCK) {
      if (dstat&BT848_DSTATUS_HLOC)
	audio(btv, AUDIO_ON);
      else
	audio(btv, AUDIO_OFF);
    }
    
    if (astat&BT848_INT_I2CDONE) {
    }
    
    count++;
    
#ifdef DEBUG
    if (count>1)
      printk ("bttv: irq loop %d\n", count);
#endif
      
    if (count > 10) {
      btwrite(0, BT848_INT_MASK);
      printk ("bttv: IRQ lockup, cleared int mask\n");
    }
  }
}


/* scan for a Bt848 card, request the irq and map the io memory */
static int
find_bt848(void)
{
  short pci_index;    
  unchar command, latency;
  int result;
  unchar bus, devfn;
  struct bttv *btv;
  char st[256];

  bttv_num=0;

  if (!pcibios_present()) {
    printk("bttv: PCI-BIOS not present or not accessable!\n");
    return 0;
  }

  for (pci_index = 0;
       !pcibios_find_device(PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT848,
			    pci_index, &bus, &devfn);
       ++pci_index) {
    btv=&bttvs[bttv_num];
    btv->bus=bus;
    btv->devfn=devfn;
    btv->bt848_mem=NULL;
    btv->vbibuf=NULL;
    btv->risc_jmp=NULL;
    btv->vbi_odd=NULL;
    btv->vbi_even=NULL;
    btv->vbiq=NULL;
    btv->capq=NULL;
    btv->vbip=VBIBUF_SIZE;
    
    pcibios_read_config_byte(btv->bus, btv->devfn,
			     PCI_INTERRUPT_LINE, &btv->irq);
    pcibios_read_config_dword(btv->bus, btv->devfn, PCI_BASE_ADDRESS_0,
			      &btv->bt848_adr);
    
    if (remap&&(!bttv_num)) { 
      remap<<=20;
      remap&=PCI_BASE_ADDRESS_MEM_MASK;
      printk("remapping to : 0x%08x.\n", remap);
      remap|=btv->bt848_adr&(~PCI_BASE_ADDRESS_MEM_MASK);
      pcibios_write_config_dword(btv->bus, btv->devfn, PCI_BASE_ADDRESS_0,
				 remap);
      pcibios_read_config_dword(btv->bus, btv->devfn, PCI_BASE_ADDRESS_0,
				&btv->bt848_adr);
    }
    
    btv->bt848_adr&=PCI_BASE_ADDRESS_MEM_MASK;
    pcibios_read_config_byte(btv->bus, btv->devfn, PCI_CLASS_REVISION,
			     &btv->revision);
    printk("bttv: Brooktree Bt848 (revision %d)\n",btv->revision);
    printk("bttv: bus: %d, devfn: %d, ",
	   btv->bus, btv->devfn);
    printk("irq: %d, ",btv->irq);
    printk("memory: 0x%08x.\n", btv->bt848_adr);
    
    btv->bt848_mem=ioremap(btv->bt848_adr, 0x1000);

    result = request_irq(btv->irq, bttv_irq,
			 SA_SHIRQ | SA_INTERRUPT,"bttv",(void *)btv);
    if (result==-EINVAL) {
      printk("bttv: Bad irq number or handler\n");
      return -EINVAL;
    }
    if (result==-EBUSY) {
      printk("bttv: IRQ %d busy, change your PnP config in BIOS\n",btv->irq);
      return result;
    }
    if (result < 0) 
      return result;

/*
    if (result==-EBUSY) {
      for (i=0; i<bttv_num; i++)
	if (bttvs[i].irq==btv->irq)
	  goto shared;
      printk("IRQ %d busy\n",btv->irq);
      return -EBUSY;
    shared:
      printk("bttv: Sharing IRQ %d with other Bt848 card(s)\n",btv->irq);
    }      
*/

    /* enable bus-mastering */
    pcibios_read_config_byte(btv->bus, btv->devfn, PCI_COMMAND, &command);
    command|=PCI_COMMAND_MASTER;
    pcibios_write_config_byte(btv->bus, btv->devfn, PCI_COMMAND, command);
    pcibios_read_config_byte(btv->bus, btv->devfn, PCI_COMMAND, &command);
    if (!(command&PCI_COMMAND_MASTER)) {
      printk("bttv: PCI bus-mastering could not be enabled.\n");
      return -1;
    }
    pcibios_read_config_byte(btv->bus, btv->devfn, PCI_LATENCY_TIMER, 
			     &latency);
    if (!latency) {
      latency=32;
      pcibios_write_config_byte(btv->bus, btv->devfn, PCI_LATENCY_TIMER,
				latency);
    }
    printk("bttv: latency: 0x%02x\n", latency);
    
    if (check_region(btv->bt848_adr,256))
       {
        sprintf(st,"bttv%d",bttv_num);
        if (!(request_region(btv->bt848_adr,256,st)))
           {
            printk("bttv: WARNING: request region failed\n");
           }  	
       }
    
    bttv_num++;
  }
  printk("bttv: %d Bt848 card(s) found.\n", bttv_num);
  return bttv_num;
}

static void
release_bttv()
{
  unchar command;
  int i;
  struct bttv *btv;

  for (i=0;i<bttv_num; i++) {
    btv=&bttvs[i];
    /* turn off all capturing, DMA and IRQs */

    /* first disable interrupts before unmapping the memory! */
    btwrite(0, BT848_INT_MASK);
    btwrite(0xffffffffUL,BT848_INT_STAT);
    btwrite(0x0, BT848_GPIO_OUT_EN);

    bt848_cap(btv, 0);
    
    /* disable PCI bus-mastering */
    pcibios_read_config_byte(btv->bus, btv->devfn, PCI_COMMAND, &command);
    command|=PCI_COMMAND_MASTER;
    pcibios_write_config_byte(btv->bus, btv->devfn, PCI_COMMAND, command);
    
    /* unmap and free memory */
    if (btv->risc_odd)
      kfree((void *) btv->risc_odd);
    if (btv->risc_even)
      kfree((void *) btv->risc_even);

    printk("free: risc_jmp: 0x%p.\n", btv->risc_jmp);
    if (btv->risc_jmp)
      kfree((void *) btv->risc_jmp);

    printk("bt848_vbibuf: 0x%p.\n", btv->vbibuf);
    if (btv->vbibuf)
      vfree((void *) btv->vbibuf);
    if (btv->cliprecs)
      vfree((void *) btv->cliprecs);
    free_irq(btv->irq,btv);
    printk("bt848_mem: 0x%p.\n", btv->bt848_mem);
    if (btv->bt848_mem)
      iounmap(btv->bt848_mem);
      
    release_region(btv->bt848_adr,256);  
  }
}



#ifdef __cplusplus
extern "C" {
#endif

extern int printk(const char* fmt, ...);

int
init_module( void)
{
  int i;

  handle_chipset();

  if (find_bt848()<0)
    return -EIO;

  for (i=0; i<bttv_num; i++) 
    if (init_bt848(&bttvs[i])<0) {
      release_bttv();
      return -EIO;
    } 
  
  if (register_chrdev(major, "bttv", &bttv_fops)) 
     {
      printk("register_chrdev failed.\n");
      release_bttv();
      return -EIO;
     }

  return 0;
}

void
cleanup_module(void)
{
  release_bttv();
  
  if (unregister_chrdev(major, "bttv") != 0)
    printk("bttv: cleanup_module failed\n");
  else
    printk("bttv: signing off\n");
}

#ifdef __cplusplus
}
#endif
